try:
	from .pyllrp import *
	from . import TagInventory
except Exception as e:
	from pyllrp import *
	import TagInventory

def padToWords( epc ):
	epc = '{}'.format(epc)
	len_epc = len(epc)
	return epc.zfill( len_epc + (-len_epc%4) )
	
def getTagMask( epc ):
	epc = padToWords( epc )
	return '{:04X}{}'.format( ((1<<5)-1)<<(16-5), 'F' * len(epc) )
	
def addLengthPrefix( epc ):
	# Pad out to a 16-bit word.
	# Then, add the length in words as the highest 5-bits of the first word.
	# We don't care about the other bits - just leave them at zero.
	epc = padToWords( epc )
	epc = '{:04X}{}'.format( (len(epc)//4)<<(16-5), epc )
	return epc
	
def hexToWords( epc ):
	assert len(epc) % 4 == 0, 'epc must be a 16-bit word multiple'
	return [int(epc[i:i+4], 16) for i in range(0, len(epc), 4)]

def hexToBytes( epc ):
	assert len(epc) % 2 == 0, 'epc must be a byte multiple'
	return codec.decode(epc, 'hex_codec')

class TagWriter( TagInventory.TagInventory ):
	accessSpecID = 456

	def GetAccessSpec(	self,
						MessageID = None,			# If None, one will be assigned.
						epcOld = '',				# Old EPC.  Empty matches anything.
						epcNew = '0123456789',		# New EPC.
						roSpecID = 0,				# ROSpec to trigger this ROSpec.  0 = run when any ROSpec runs.
						opSpecID = 1,				# Something unique.
						operationCount = 10,		# Number of times to execute.
						antenna = None ):
		
		TagMask = hexToBytes( getTagMask(epcOld) ) if epcOld else b''
		TagData = hexToBytes( addLengthPrefix(epcOld) ) if epcOld else b''
		
		accessSpecMessage = ADD_ACCESSSPEC_Message( MessageID = MessageID, Parameters = [
			AccessSpec_Parameter(
				AccessSpecID  = self.accessSpecID,
				AntennaID = antenna if antenna else 0,	# Unspecified antenna: apply to all antennas
				ProtocolID = AirProtocols.EPCGlobalClass1Gen2,
				CurrentState = bool(AccessSpecState.Disabled),
				ROSpecID = roSpecID,
				
				Parameters = [
					AccessSpecStopTrigger_Parameter(
						AccessSpecStopTrigger = AccessSpecStopTriggerType.Operation_Count,
						OperationCountValue = operationCount
					),
					AccessCommand_Parameter(
						Parameters = [
							C1G2TagSpec_Parameter(
								Parameters = [
									C1G2TargetTag_Parameter(
										MB = 1,						# Memory Bank 1 = EPC
										Pointer = 16,				# 16 bits offset - skip CRC.
										TagMask = TagMask,
										TagData = TagData,
										Match = True,
									)
								]
							),
							C1G2Write_Parameter(
								MB = 1,
								WordPointer = 1,					# Skip CRC, but include the length and flags.
								WriteData = hexToWords( addLengthPrefix(epcNew) ),
								OpSpecID = opSpecID,
								AccessPassword = 0,
							),
						]
					),
					AccessReportSpec_Parameter(
						AccessReportTrigger = AccessReportTriggerType.End_Of_AccessSpec
					)
				]
			)
		])	# ADD_ACCESS_SPEC_Message
		
		return accessSpecMessage
		
	def _prologTW( self, epcOld, epcNew, antenna = None ):
		response = self.connector.transact( DELETE_ACCESSSPEC_Message(AccessSpecID = 0) )
		assert response.success(), 'Delete AccessSpec Fails\n{}'.format(response)
		
		message = self.GetAccessSpec(
					epcOld = epcOld,
					epcNew = epcNew,
					antenna = antenna,
		)
		response = self.connector.transact( message )
		assert response.success(), 'Add AccessSpec Fails\n{}'.format(response)
	
	def _executeTW( self ):
		response = self.connector.transact( ENABLE_ACCESSSPEC_Message(AccessSpecID = self.accessSpecID) )
		assert response.success(), 'Enable AccessSpec Fails\n{}'.format(response)
		
		# Run the TagInventory to trigger our AccessSpec.
		self._execute()
		
		response = self.connector.transact( DISABLE_ACCESSSPEC_Message(AccessSpecID = self.accessSpecID) )
		# What to do if the Enable succeeds but the Disable fails?
		# This should be OK as we always delete and recreate the message between calls.
		#assert response.success(), 'Disable AccessSpec Fails\n{}'.format(response)
		
	def _epilogTW( self ):
		response = self.connector.transact( DELETE_ACCESSSPEC_Message(AccessSpecID = self.accessSpecID) )
		assert response.success(), 'Delete AccessSpec Fails\n{}'.format(response)
		
	def WriteTag( self, epcOld, epcNew, antenna = None ):
		''' Change a single tag. '''
		self._prolog( antenna )
		self._prologTW( epcOld, epcNew, antenna )
		
		self._executeTW()
		
		self._epilogTW()
		self._epilog()
		
	def WriteFOVConsecutive( self, tagStart, tagFormat = '{}' ):
		''' Change all tags in the "Field Of View" to consecutive values, starting from tagStart. '''
		tagInventory, otherMessages = self.GetTagInventory()
		
		self._prolog()
		tagCur = tagStart
		for tag in sorted(tagInventory, reverse=True):	# Go in reverse to avoid overlap increment conflicts.
			self._prologTW( tag, tagFormat.format(tagStart) )
			self._executeTW()
			self._epilogTW()
			tagCur += 1
			
		self._epilog()
		return tagCur

if __name__ == '__main__':
	print ( '{}\n'.format('*' * 75) )
	
	impinjHost = '192.168.0.101'
	tw = TagWriter( impinjHost )
	tw.Connect()
	
	print ( 'Before tags' )
	tagInventory, otherMessages = tw.GetTagInventory()
	print ( '\n'.join( tagInventory ) )
	
	tagValue = '256'
	print ( 'Writing tag: {}'.format(tagValue) )
	tw.WriteTag( '', tagValue )
	
	print ( 'After Tags' )
	tagInventory, otherMessages = tw.GetTagInventory()
	print ( '\n'.join( tagInventory ) )
	
	tw.Disconnect()
	tw.Connect()
	tagInventory, otherMessages = tw.GetTagInventory()
	for tag in tagInventory:
		tw.Connect()
		epcOld = tag
		epcNew = u'{:X}'.format(int(tag,16)+1)
		print ( u'{} {}'.format(epcOld, epcNew) )
		tw.WriteTag( epcOld, epcNew )
		tw.Disconnect()

	tw.Connect()
	print ( '{}\n'.format('*' * 75) )
	tagInventory, otherMessages = tw.GetTagInventory()
	print ( '\n'.join(tagInventory) )
	
	tw.Disconnect()
	



#---------------------------------------------------------------------------
#
# Reformats the llrpdef.xml file into a Python-formatted description llrpdef.py
#
# The reason we so this is so we avoid a dependency on the xml parser in deployment.
# It is also much faster to use the .pyc file rather than parsing XML again
# on startup.
#
import sys
import pprint
from xml.dom.minidom import parse

# Map the xml field types to our types (bitstring standard, and our custom ones).
fieldMap = {
	'u1':		'bool',
	'u2':		'bits:2',
	'u8':		'uintbe:8',
	'u16':		'uintbe:16',
	'u32':		'uintbe:32',
	'u64':		'uintbe:64',
	'u96':		'uintbe:96',
	
	's8':		'intbe:8',
	's16':		'intbe:16',
	's32':		'intbe:32',
	's64':		'intbe:64',
	
	'utf8v':	'string',
	
	'u1v':		'bitarray',
	'u8v':		'array:8',
	'u16v':		'array:16',
	'u32v':		'array:32',
	
	'bytesToEnd':	'bytesToEnd',	# Used for the Custom message type.
}

def toAscii( s ):
	return s.encode('ascii', 'ignore').decode( 'ascii' )

def getEnum( e ):
	Name = toAscii(e.attributes['name'].value)
	Choices = [ (int(toAscii(c.attributes['value'].value)), toAscii(c.attributes['name'].value))
		for c in e.childNodes if c.nodeName == 'entry' ]
	return {'name':Name, 'choices':tuple(Choices) }

def getParameterMessage( n, isMessage ):
	Name = toAscii(n.attributes['name'].value)
	
	Fields = []
	Parameters = []
	
	sys.stdout.write( '{} {}\n'.format(Name, 'Message' if isMessage else 'Parameter') )
	try:
		TypeNum = int(n.attributes['typeNum'].value)
	except KeyError:
		# This is a custom parameter or message.
		TypeNum = 1023		# Code for custom Message and Parameter.
		Fields.append( {'name': 'VendorIdentifier', 'type': 'uintbe:32', 'default': vendors[toAscii(n.attributes['vendor'].value)]} )
		if isMessage:
			Fields.append( {'name': 'MessageSubtype', 'type': 'uintbe:8', 'default': int(n.attributes['subtype'].value)} )
		else:
			Fields.append( {'name': 'ParameterSubtype', 'type': 'uintbe:32', 'default': int(n.attributes['subtype'].value)} )
	
	for c in n.childNodes:
		if c.nodeName == 'field':
			fieldInfo = { 'name': toAscii(c.attributes['name'].value), 'type': fieldMap[toAscii(c.attributes['type'].value)] }
			try:
				fieldInfo['enumeration'] = toAscii(c.attributes['enumeration'].value)
			except Exception:
				pass
			try:
				fieldInfo['format'] = toAscii(c.attributes['format'].value)
			except Exception:
				pass
			Fields.append( fieldInfo )
		elif c.nodeName == 'reserved':
			bitCount = int(toAscii(c.attributes['bitCount'].value))
			code = 'skip:%d' % bitCount
			Fields.append( {'name':code, 'type':code} )
		elif c.nodeName == 'parameter' or c.nodeName == 'choice':
			repeat = toAscii(c.attributes['repeat'].value)
			minMax = repeat.split('-')
			if len(minMax) == 1:
				rMin = rMax = int(repeat)
			else:
				rMin = int(minMax[0])
				rMax = int(minMax[1]) if minMax[1] != 'N' else 99999
			Parameter = toAscii(c.attributes['type'].value)
			Parameters.append( {'parameter':Parameter, 'repeat': (rMin, rMax), 'choice': int(c.nodeName == 'choice')} )
	pm = {'name':Name, 'typeNum':TypeNum}
	if Fields:
		pm['fields'] = tuple(Fields)
	if Parameters:
		pm['parameters'] = tuple(Parameters)
	return pm
	
def getVendorCode( n ):
	name = toAscii(n.attributes['name'].value)
	id = int(n.attributes['vendorID'].value)
	return (name, id)

def getChoiceDefinition( n ):
	choices = {}
	name = toAscii(n.attributes['name'].value)
	for c in n.childNodes:
		if c.nodeName == 'parameter':
			choices[toAscii(c.attributes['type'].value)] = name	
	return choices

enums, parameters, messages = [], [], []
vendors = {}
choiceDefinitions = {}

llrpDefXml = '../xml/llrp-1x0-def.xml'
dom = parse( llrpDefXml )

vendors.update( dict(getVendorCode(v) for v in dom.getElementsByTagName('vendorDefinition')) )
enums = [getEnum(e) for e in dom.getElementsByTagName('enumerationDefinition')]
parameters = [getParameterMessage(p, False) for p in dom.getElementsByTagName('parameterDefinition')]
messages = [getParameterMessage(m, True) for m in dom.getElementsByTagName('messageDefinition')]
for c in dom.getElementsByTagName('choiceDefinition'):
	choiceDefinitions.update( getChoiceDefinition(c) )
del choiceDefinitions['Custom']

dom.unlink()
dom = None

llrpCustomDefXml = '../xml/Impinjdef-1.18-private.xml'
dom = parse( llrpCustomDefXml )

vendors.update( dict(getVendorCode(v) for v in dom.getElementsByTagName('vendorDefinition')) )
enums.extend( getEnum(e) for e in dom.getElementsByTagName('customEnumerationDefinition') )
parameters.extend( getParameterMessage(p, False) for p in dom.getElementsByTagName('customParameterDefinition') )
messages.extend( getParameterMessage(m, True) for m in dom.getElementsByTagName('customMessageDefinition') )
for p in dom.getElementsByTagName('customParameterDefinition'):
	choiceDefinitions[toAscii(p.attributes['name'].value)] = 'Custom'

with open('llrpdef.py', 'w', encoding='utf8') as fp:
	fp.write( '#-----------------------------------------------------------\n' )
	fp.write( '# DO NOT EDIT!\n' )
	fp.write( '# MACHINE GENERATED by ParseDef.py\n' )
	fp.write( '# XML sources:\n' )
	fp.write( '#    "{}"\n'.format(llrpDefXml) )
	fp.write( '#    "{}"\n'.format(llrpCustomDefXml) )
	fp.write( '# see https://github.com/esitarski/pyllrp for details\n' )
	fp.write( '#-----------------------------------------------------------\n' )
	for n in ['vendors', 'enums', 'parameters', 'messages', 'choiceDefinitions']:
		fp.write( '{}='.format(n) )
		pprint.pprint( globals()[n], fp, compact=True )
		fp.write( '\n' )



#!/usr/bin/env python

import os
import sys
import stat
import glob
import shutil
import datetime
import subprocess

import ParseDef

version = '3.0.0'

pypiDir = 'pypi'

def removeTabs( buf, tabStop = 4 ):
	# Remove tabs from Python code and preserve formatting.
	lines = []
	for line in buf.split( '\n' ):
		lineOut = []
		for c in line:
			if c == '\t':
				lineOut.append( ' ' )
				while len(lineOut) % tabStop != 0:
					lineOut.append( ' ' )
			else:
				lineOut.append( c )
		lines.append( ''.join(lineOut) )
	return '\n'.join( lines ) + '\n'

def writeToFile( s, fname ):
	print( 'creating', fname, '...' )
	with open(os.path.join(pypiDir,fname), 'w', encoding='utf8') as f:
		f.write( s )

#------------------------------------------------------
# Create a release area for pypi
print( 'Clearing previous contents...' )
try:
	subprocess.call( ['rm', '-rf', pypiDir] )
except:
	try:
		shutil.rmtree( pypiDir, ignore_errors=True )
	except:
		pass
	
os.mkdir( pypiDir )

#--------------------------------------------------------
readme = '''
======
pyllrp
======

A pure Python implementation of LLRP (Low Level Reader Protocol)
used by RFID readers including Impinj, Alien and ThingMagic.
See the LLRP spec for details on the messages.

Allows quick-and-easy scripting in fully portable pure Python to create LLRP applications.

All Message and Parameters are full Python classes.
Full validation of all LLRP Messages and Parameters including data type, date values and parameter checking.
Full support for enumerated values.

See TinyExample.py for how to use.

A reader connection manager is also included that can connect to a reader, transact commands, then start/stop a thread to listen for tag reads.  A message handler can be configured to respond to any reader message.  See wxExample.py for a simple method to show reader messages in a wxPython application with a Queue (requires wxPython install).

The module also supports reading and writing messages in XML format.
'''

writeToFile( readme, 'README.txt' )

#--------------------------------------------------------
changes = '''
'''
writeToFile( changes, 'CHANGES.txt' )
	
#--------------------------------------------------------
license = '''
Copyright (C) 2013-%d Edward Sitarski

Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
and associated documentation files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
''' % datetime.datetime.now().year

writeToFile( license, 'License.txt' )

#--------------------------------------------------------

manifest = '''include *.txt
'''

writeToFile( manifest, 'MANIFEST.in' )

#--------------------------------------------------------

srcDir = os.path.join( pypiDir, 'pyllrp' )
os.mkdir( srcDir )

print( 'Copy the src files and add the copyright notice.' )
license = license.replace( '\n', '\n# ' )
for fname in glob.glob( '*.*' ):
	if not (fname.endswith( '.py' ) or fname.endswith('.pyw')) or fname == 'pypi.py':
		continue
	print( '   ', fname, '...' )
	with open(fname, 'r', encoding='utf8') as f:
		contents = f.read()
	if contents.startswith('import'):
		p = 0
	else:
		for search in ['\nimport', '\nfrom']:
			p = contents.find(search)
			if p >= 0:
				p += 1
				break
	if p >= 0:
		contents = ''.join( [contents[:p],
							'\n',
							'#------------------------------------------------------',
							license,
							'\n',
							contents[p:]] )
							
	contents = removeTabs( contents )
	contents.replace( '\r\n', '\n' )
	with open(os.path.join(srcDir, fname, encoding='utf8'), 'w' ) as f:
		f.write( contents )


print( 'Creating setup.py...' )
setup = {
	'name':			'pyllrp',
	'version':		version,
	'author':		'Edward Sitarski',
	'author_email':	'edward.sitarski@gmail.com',
	'url':			'http://www.sites.google.com/site/crossmgrsoftware/',
	'packages':		['pyllrp'],
	'license':		'License.txt',
	'include_package_data': True,
	'description':	'pyllrp: a pure Python implementation of LLRP (Low Level Reader Protocol).',
	'install_requires':	[
		'bitstring >= 3.1.1',
	],
	'classifiers' : [
		'Development Status :: 3 - Alpha',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3',
		"Operating System :: OS Independent",
	],
}

with open(os.path.join(pypiDir,'setup.py', encoding='utf8'), 'w') as f:
	f.write( 'from distutils.core import setup\n' )
	f.write( 'setup(\n' )
	for key, value in setup.items():
		f.write( '    {}={},\n'.format(key, repr(value)) )
	f.write( "    long_description=open('README.txt').read(),\n" )
	f.write( ')\n' )

print( 'Creating install package...' )
os.chdir( pypiDir )
subprocess.call( ['python', 'setup.py', 'sdist', '--formats=gztar,zip'] )

os.chdir( 'dist' )
shutil.move( 'pyllrp-{}.zip'.format(version), 'pip-install-pyllrp-{}.zip'.format(version) )
shutil.move( 'pyllrp-{}.tar.gz'.format(version), 'pip-install-pyllrp-{}.tar.gz'.format(version) )
	
print( 'Done.' )



#!/usr/bin/env python
import time
import socket
try:
	from .pyllrp import *
	from .LLRPConnector import LLRPConnector
except Exception as e:
	from pyllrp import *
	from LLRPConnector import LLRPConnector

class TagInventory:
	roSpecID = 123					# Arbitrary roSpecID.
	inventoryParameterSpecID = 1234	# Arbitrary inventory parameter spec id.
	readWaitMilliseconds = 100

	def __init__( self, host='192.168.10.102', defaultAntennas=None, transmitPower=None, receiverSensitivity=None ):
		self.host = host
		self.connector = None
		self.resetTagInventory()
		self.defaultAntennas = defaultAntennas
		self.transmitPower = transmitPower
		self.receiverSensitivity = receiverSensitivity
		
	def resetTagInventory( self ):
		self.tagInventory = set()
		self.tagDetail = []
		self.otherMessages = []

	def AccessReportHandler( self, connector, accessReport ):
		for tag in accessReport.getTagData():
			tag['Tag'] = HexFormatToStr( tag['EPC'] )
			self.tagInventory.add( tag['Tag'] )
			self.tagDetail.append( tag )

	def DefaultHandler( self, connector, message ):
		self.otherMessages.append( message )

	def _getReaderConfigMessage( self ):
		# Change receiver sensitivity (if specified).  This value is RFID reader dependent.
		receiverSensitivityParameter = []
		if self.receiverSensitivity is not None:
			receiverSensitivityParameter.append(
				RFReceiver_Parameter( 
					ReceiverSensitivity = self.receiverSensitivity
				)
			)
		
		# Change transmit power (if specified).  This value is RFID reader dependent.
		transmitPowerParameter = []
		if self.transmitPower is not None:
			transmitPowerParameter.append(
				RFTransmitter_Parameter( 
					TransmitPower = self.transmitPower,
					HopTableID = 1,						# Ignored in non-frequency-hopping regulatory regions. 
					ChannelIndex = 1,					# Ignored in non-frequency-hopping regulatory regions.
				)
			)
		
		return SET_READER_CONFIG_Message( Parameters = [
			AntennaConfiguration_Parameter( AntennaID = 0, Parameters =
				receiverSensitivityParameter +
				transmitPowerParameter + [
				C1G2InventoryCommand_Parameter( Parameters = [
					C1G2SingulationControl_Parameter(
						Session = 0,
						TagPopulation = 100,
						TagTransitTime = 3000,
					),
				] ),
			] ),
		] )

	def Connect( self ):
		# Create the reader connection.
		self.connector = LLRPConnector()

		# Connect to the reader.
		try:
			response = self.connector.connect( self.host )
		except socket.timeout:
			self.connector.disconnect()
			raise
			
		# Reset to factory defaults.
		response = self.connector.transact( SET_READER_CONFIG_Message(ResetToFactoryDefault = True) )
		assert response.success(), f'SET_READER_CONFIG ResetToFactorDefault fails\n{response}'

		message = self._getReaderConfigMessage()
		response = self.connector.transact( message )
		assert response.success(), 'SET_READER_CONFIG Configuration fails:\n{}'.format(response)
		
	def Disconnect( self ):
		self.connector.disconnect()
		self.connector = None

	def GetROSpec( self, antennas = None ):
		if antennas is not None:
			if not isinstance(antennas, list):
				antennas = [antennas]
		else:
			antennas = [0]
	
		# Create an rospec that reports reads.
		return ADD_ROSPEC_Message( Parameters = [
				ROSpec_Parameter(
					ROSpecID = self.roSpecID,
					CurrentState = ROSpecState.Disabled,
					Parameters = [
						ROBoundarySpec_Parameter(		# Configure boundary spec (start and stop triggers for the reader).
							Parameters = [
								ROSpecStartTrigger_Parameter(ROSpecStartTriggerType = ROSpecStartTriggerType.Immediate),
								ROSpecStopTrigger_Parameter(ROSpecStopTriggerType = ROSpecStopTriggerType.Null),
							]
						), # ROBoundarySpec
						AISpec_Parameter(				# Antenna Inventory Spec (specifies which antennas and protocol to use)
							AntennaIDs = antennas,
							Parameters = [
								AISpecStopTrigger_Parameter(
									AISpecStopTriggerType = AISpecStopTriggerType.Tag_Observation,
									Parameters = [
										TagObservationTrigger_Parameter(
											TriggerType = TagObservationTriggerType.N_Attempts_To_See_All_Tags_In_FOV_Or_Timeout,
											NumberOfAttempts = 10,
											Timeout = self.readWaitMilliseconds,
										),
									]
								),
								InventoryParameterSpec_Parameter(
									InventoryParameterSpecID = self.inventoryParameterSpecID,
									ProtocolID = AirProtocols.EPCGlobalClass1Gen2,
								),
							]
						), # AISpec
						ROReportSpec_Parameter(			# Report spec (specifies how often and what to send from the reader)
							ROReportTrigger = ROReportTriggerType.Upon_N_Tags_Or_End_Of_ROSpec,
							N = 10000,
							Parameters = [
								TagReportContentSelector_Parameter(
									EnableAntennaID = True,
									EnableFirstSeenTimestamp = True,
									EnablePeakRSSI = True,
								),
							]
						), # ROReportSpec
					]
				), # ROSpec_Parameter
			]
		)	# ADD_ROSPEC_Message

	def _prolog( self, antennas = None ):
		# Disable all the rospecs.  This command may fail so we ignore the response.
		response = self.connector.transact( DISABLE_ROSPEC_Message(ROSpecID = 0) )
		# Delete our old rospec if it exists.  This command might fail so we ignore the response.
		response = self.connector.transact( DELETE_ROSPEC_Message(ROSpecID = self.roSpecID) )

		# Add callbacks so we can record the tag reads and any other messages from the reader.
		self.resetTagInventory()
		self.connector.addHandler( RO_ACCESS_REPORT_Message, self.AccessReportHandler )
		self.connector.addHandler( 'default', self.DefaultHandler )

		# Add and enable our ROSpec
		response = self.connector.transact( self.GetROSpec(antennas) )
		assert response.success(), 'Add ROSpec Fails\n{}'.format(response)
		
	def _execute( self ):
		response = self.connector.transact( ENABLE_ROSPEC_Message(ROSpecID = self.roSpecID) )
		assert response.success(), 'Enable ROSpec Fails\n{}'.format(response)
		
		# Wait for the reader to do its work.
		time.sleep( (1.5*self.readWaitMilliseconds) / 1000.0 )
		
		response = self.connector.transact( DISABLE_ROSPEC_Message(ROSpecID = self.roSpecID) )
		assert response.success(), 'Disable ROSpec Fails\n{}'.format(response)
		
	def _epilog( self ):
		# Cleanup.
		response = self.connector.transact( DELETE_ROSPEC_Message(ROSpecID = self.roSpecID) )
		assert response.success(), 'Delete ROSpec Fails\n{}'.format(response)
		self.connector.removeAllHandlers()
		
	def GetTagInventory( self, antennas = None ):
		self._prolog( antennas if antennas is not None else self.defaultAntennas )
		self._execute()
		self._epilog()
		return self.tagInventory, self.otherMessages

if __name__ == '__main__':
	'''Read a tag inventory from the reader and shutdown.'''
	print( '**** full power' )
	host = '192.168.0.101'
	ti = TagInventory( host )
	ti.Connect()
	tagInventory, otherMessages = ti.GetTagInventory()
	print( '\n'.join(tagInventory) )
	ti.Disconnect()
	
	for p in range(1, 100, 10):
		ti = TagInventory( host, transmitPower = p )
		ti.Connect()
		tagInventory, otherMessages = ti.GetTagInventory()
		print( f'**** power={p}' )
		print( '\n'.join(tagInventory) )
		ti.Disconnect()



#!/usr/bin/env python
from queue import Queue, Empty
import socket
import datetime
import threading
try:
	from .pyllrp import *
except Exception as e:
	from pyllrp import *

class LLRPConnector:
	#--------------------------------------------------------------------------
	#
	# A simple LLRP reader connection manager.
	#
	# Supports connecting to the reader, transacting commands, message handlers,
	# and asynchronous/synchronous monitoring of the reader socket.
	#
	def __init__( self ):
		self.TimeoutSecs = 6		# Time for the reader to respond.
		self._reset()
		self.handlers = {}
		
	def _reset( self ):
		''' Reset all internal fields. '''
		self.host = None
		self.port = None
		self.readerSocket = None
		self.thread = None
		self.shutdownQ = None		# Used to shutdown the thread.
		self.keepGoing = False
		self.timeCorrection = None
	
	def connect( self, host, port = 5084 ):
		''' Connect to a reader. '''
		self._reset()
		self.readerSocket = socket.socket( socket.AF_INET, socket.SOCK_STREAM )
		
		# Set a timeout for the socket.  This is also the maximum time it will take to shut down the listener.
		self.readerSocket.settimeout( self.TimeoutSecs )
		self.readerSocket.connect( (host, port) )
		
		self.host = host
		self.port = port
		
		# Expecting READER_EVENT_NOTIFICATION message.
		response = UnpackMessageFromSocket( self.readerSocket )
		tNow = datetime.datetime.now()	# Get the time here to minimize latency.
		
		# Check if the connection succeeded.
		connectionAttemptEvent = response.getFirstParameterByClass(ConnectionAttemptEvent_Parameter)
		if connectionAttemptEvent and connectionAttemptEvent.Status != ConnectionAttemptStatusType.Success:
			self.disconnect()
			raise EnvironmentError(
				connectionAttemptEvent.Status,
				ConnectionAttemptStatusType.getName(connectionAttemptEvent.Status).replace('_',' ')
			)
		
		self.keepGoing = True
		
		# Compute a correction between the reader's time and the computer's time.
		self.timeCorrection = None
		try:
			microseconds = response.getFirstParameterByClass(UTCTimestamp_Parameter).Microseconds
			readerTime = datetime.datetime.utcfromtimestamp( microseconds / 1000000.0 )
			self.timeCorrection = tNow - readerTime
		except Exception as e:
			self.disconnect()
			raise ValueError('Missing Timestamp: ' + response.__repr__())

		return response
		
	def tagTimeToComputerTime( self, tagTime ):
		# Time is in microseconds from Jan 1, 1970.
		return datetime.datetime.utcfromtimestamp( tagTime / 1000000.0 ) + self.timeCorrection
	
	def disconnect( self ):
		''' Disconnect from a reader.  Also stops the listener. '''
		self.timeCorrection = None
		if not self.readerSocket:
			return None
		
		if self.isListening():
			self.stopListener()
		
		# Send the reader a disconnect message.
		response = None
		try:
			response = self.transact( CLOSE_CONNECTION_Message() )
		except Exception:
			pass
		
		self.readerSocket.close()
		self.readerSocket = None
		return response
		
	def addHandler( self, messageClass, handlerFunc ):
		''' Add a handler for a specific message type. '''
		''' Support multiple handlers for the same message type. '''
		self.handlers.setdefault( messageClass, [] ).append( handlerFunc )
		
	def removeHandler( self, messageClass, handlerFunc = None ):
		''' Remove a handler for a specific message type from the reader.
			If handlerFunc is None, all handlers for the given messageClass will be removed.
			If handlerFunc is not None, only the specific handler is removed.
		'''
		if handlerFunc is None:
			try:
				del self.handlers[messageClass]
			except KeyError:
				pass
		else:
			while True:
				try:
					self.handlers[messageClass].remove( handlerFunc )
				except (KeyError, ValueError):
					break
					
	def removeAllHandlers( self ):
		self.handlers = {}
	
	def transact( self, message ):
		''' Send a message to the reader and wait for the response. '''
		assert not self.isListening(), 'Cannot perform transact() while listen thread is active.  Stop it first with stopListener().'
		message.send( self.readerSocket )
		response = WaitForMessage( message.MessageID, self.readerSocket, self.callHandler )
		return response
		
	def checkKeepGoing( self ):
		''' Check if we should continue the reader thread. '''
		if not self.keepGoing:
			return False
			
		try:
			# Check the shutdown queue for a message.  If there is one, shutdown.
			self.shutdownQ.get( False )
			self.shutdownQ.task_done()
			self.keepGoing = False
			return False
		except (AttributeError, Empty):
			return True
	
	def callHandler( self, message ):
		''' Call all the handlers for this message. '''
		for cb in (self.handlers.get(message.__class__, None) or self.handlers.get('default', [])):
			cb( self, message )
	
	def listen( self ):
		''' Listen for messages from the reader. '''
		# Calling this by itself will block.
		# Recommended usage is to use startListener and stopListener.
		while self.checkKeepGoing():
			try:
				response = UnpackMessageFromSocket( self.readerSocket )
			except socket.timeout:
				continue
			self.callHandler( response )
				
	def startListener( self ):
		''' Starts a thread to listen to the reader. '''
		assert self.readerSocket, 'Cannot start listener without a successful connection.'
		assert not self.thread, 'Listener is already running.  Stop it first with stopListener().'
		self.shutdownQ = Queue()
		self.keepGoing = True
		self.thread = threading.Thread( target = self.listen, name='LLRP Listener' )
		self.thread.daemon = True
		self.thread.start()
	
	def stopListener( self ):
		''' Stops the thread listening to the reader. '''
		self.shutdownQ.put( 'Shutdown' )
		self.thread.join()					# Wait for the thread to terminate.
		self.shutdownQ = None
		self.thread = None
	
	def isListening( self ):
		return self.thread and self.thread.is_alive()
		
	def __enter__( self ):
		return self
		
	def __exit__(self, type, value, traceback):
		self.disconnect()



from distutils.core import setup
setup(
    name='pyllrp',
    version='3.1.1',
    author='Edward Sitarski',
    author_email='edward.sitarski@gmail.com',
    url='http://www.sites.google.com/site/crossmgrsoftware/',
    packages=['pyllrp'],
    license='License.txt',
    include_package_data=True,
    description='pyllrp: a pure Python implementation of LLRP (Low Level Reader Protocol).',
    install_requires=['bitstring >= 3.1.1',],
    classifiers=['Development Status :: 5 - Production/Stable', 'License :: OSI Approved :: MIT License', 'Programming Language :: Python :: 3', 'Operating System :: OS Independent'],
    long_description=open('README.md').read(),
	long_description_content_type='text/markdown',   
)
